"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ExpressionBuilder = void 0;
const counter_1 = require("./counter");
const placeholder_1 = require("./placeholder");
/**
 * An ExpressionBuilder instance builds the expression using calculated placeholders by referred attribute names and values.
 *
 * Each instance must correspond to the DynamoDB requests that using the expression.
 * If instances are used across different requests, the names and values that substitutes the placeholders may misplaced.
 */
class ExpressionBuilder {
    constructor() {
        this.valuesRefCounter = counter_1.createRefCounter();
        this.namesRefCounter = counter_1.createRefCounter();
    }
    /**
     * Returns merged expression attributes from added update expression and condition expression.
     */
    aggregate() {
        var _a, _b, _c, _d, _e, _f;
        if (this.updateExpressionFragment === undefined &&
            this.conditionExpressionFragment === undefined) {
            throw new Error("Neither update nor condition called");
        }
        const names = {
            ...(_a = this.updateExpressionFragment) === null || _a === void 0 ? void 0 : _a.expressionAttributeNames,
            ...(_b = this.conditionExpressionFragment) === null || _b === void 0 ? void 0 : _b.expressionAttributeNames,
        };
        const expressionAttributeValues = {
            ...(_c = this.updateExpressionFragment) === null || _c === void 0 ? void 0 : _c.expressionAttributeValues,
            ...(_d = this.conditionExpressionFragment) === null || _d === void 0 ? void 0 : _d.expressionAttributeValues,
        };
        return {
            updateExpression: (_e = this.updateExpressionFragment) === null || _e === void 0 ? void 0 : _e.expression,
            conditionExpression: (_f = this.conditionExpressionFragment) === null || _f === void 0 ? void 0 : _f.expression,
            expressionAttributeNames: compactObject(names),
            expressionAttributeValues,
        };
    }
    /**
     * Add update expression.
     */
    update(literals, ...placeholders) {
        const aggr = this.build(literals, ...placeholders);
        this.updateExpressionFragment = aggr;
        return aggr;
    }
    /**
     * Add condition expression.
     */
    condition(literals, ...placeholders) {
        const aggr = this.build(literals, ...placeholders);
        this.conditionExpressionFragment = aggr;
        return aggr;
    }
    /**
     * Builds an expression and expression attribute values from template string.
     *
     * @deprecated See `update`, `condition` and `aggregate`
     */
    expr(literals, ...placeholers) {
        return this.build(literals, ...placeholers);
    }
    /**
     *
     */
    build(literals, ...placeholers) {
        let expression = "";
        const expressionAttributeValues = {};
        const expressionAttributeNames = {};
        placeholers.forEach((pv, idx) => {
            if (placeholder_1.isAttrName(pv)) {
                const { value } = this.namesRefCounter.next();
                const placeholder = `#${value}`;
                expression += literals[idx];
                expression += placeholder;
                expressionAttributeNames[placeholder] = pv;
            }
            else {
                const { value } = this.valuesRefCounter.next();
                const ref = `:v${value}`;
                expression += literals[idx];
                expression += ref;
                expressionAttributeValues[ref] = pv;
            }
        });
        expression += literals[literals.length - 1];
        return {
            expression,
            expressionAttributeValues,
            expressionAttributeNames: compactObject(expressionAttributeNames),
        };
    }
}
exports.ExpressionBuilder = ExpressionBuilder;
const compactObject = (o) => (Object.keys(o).length === 0 ? undefined : o);
